(* on reprsente une clause par le couple constitu de la liste des
   variables positives et la liste des variables nies. *)

type clause  == (string list) * (string list);;
type systeme == clause list;;

                   (* +-------------------------+
                      |  affichage des clauses  |
                      +-------------------------+ *)

let rec concat sep = function [] -> "" | [x] -> x | x::suite -> x^sep^(concat sep suite);;
let string_of_clause(p,n) =
  (if n=[] then "" else (concat "." n)^" => ") ^
  (if p=[] then "Faux" else concat "+" p);;

let print_clause(c) = format__print_string(string_of_clause c);;
install_printer "print_clause";;


        (* +----------------------------------------------+
           |  oprations sur les systmes et les clauses  |
           +----------------------------------------------+ *)

(* compare deux clauses *)
let implique (p1,n1) (p2,n2) = (subtract p1 p2 = []) && (subtract n1 n2 = []);;

(* teste si une clause est prsente dans le systme *)
let rec present c syst = match syst with
  | [] -> false
  | x::suite -> (implique x c) || (present c suite);;

(* ajoute une clause au un systme en liminant les clauses qui en dcoulent *)
let rec ajoute c syst = match syst with
  | [] -> [c]
  | a::suite -> if implique c a
                then if implique a c then syst else ajoute c suite
                else if implique a c then syst else a :: (ajoute c suite);;

(* simplifie un systme *)
let simplifie s =
  let rec simpl s1 s2 = match s2 with
    | [] -> s1
    | c::suite -> simpl (ajoute c s1) suite
  in simpl [] s
;;

(* ajoute au systme le consensus ventuel de deux clauses *)
let ajoute_cons (p1,n1) (p2,n2) syst =
  match (intersect p1 n2, intersect p2 n1) with
    | ([x],[]) -> ajoute (union (except x p1) p2, union n1 (except x n2)) syst
    | ([],[x]) -> ajoute (union p1 (except x p2), union (except x n1) n2) syst
    | _        -> syst
;;

(* ajoute  l tous les consensus entre c et une clause du systme *)
let rec ajoute_consensus c syst l = match syst with
  | [] -> l
  | c1::suite -> ajoute_consensus c suite (ajoute_cons c c1 l)
;;

(* clt un systme par consensus *)
let rec cloture s =
  let rec consensus s1 s2 = match s1 with
    | [] -> s2
    | c::suite -> consensus suite (ajoute_consensus c s s2)
  in
  let s' = simplifie(consensus s s) in
  if (subtract s s' = []) && (subtract s' s = []) then s else cloture s'
;;

                          (* +----------+
                             |  essais  |
                             +----------+ *)


(* construction de clauses *)
let prefix =>   a b = ([b],  [a])     (* a => b         *)   
and prefix =>~  a b = ([],   [a;b])   (* a => non b     *)
and prefix ~=>  a b = ([a;b],[])      (* non a => b     *)
and prefix ~=>~ a b = ([a],  [b])     (* non a => non b *)
;;

(* devoirs et mauvaises notes *)
let syst = cloture ["d"  =>  "c"; "c"  =>~ "m"; "d" ~=>~ "m"];;

(* club cossais *)
let syst = cloture ["e" ~=>  "r";
		    "r"  =>  "k";
		    "m"  =>~ "d";
		    "d"  =>  "e"; "e" => "d";
		    "k"  =>  "e"; "k" => "m";
		    "e"  =>  "k"];;


            (* +---------------------------------------+
               |  Complment : dtail du raisonnement  |
               +---------------------------------------+ *)

type clause  == (string list) * (string list) * raison
and  raison  =  H | D of clause * clause
and  systeme == clause list;;

let string_of_clause(p,n,_) =
  (if n=[] then "" else (concat "." n)^" => ") ^
  (if p=[] then "Faux" else concat "+" p)
;;

let print_clause(c) = format__print_string(string_of_clause c);;
install_printer "print_clause";;


(* compare deux clauses *)
let implique (p1,n1,_) (p2,n2,_) = (subtract p1 p2 = []) && (subtract n1 n2 = []);;

(* teste si une clause est prsente dans le systme *)
let rec present c syst = match syst with
  | [] -> false
  | x::suite -> (implique x c) || (present c suite);;

(* ajoute une clause au un systme en liminant les clauses qui en dcoulent *)
let rec ajoute c syst = match syst with
  | [] -> [c]
  | a::suite -> if implique c a
                then if implique a c then syst else ajoute c suite
                else if implique a c then syst else a :: (ajoute c suite);;

(* simplifie un systme *)
let simplifie s =
  let rec simpl s1 s2 = match s2 with
    | [] -> s1
    | c::suite -> simpl (ajoute c s1) suite
  in simpl [] s
;;

(* ajoute au systme le consensus ventuel de deux clauses *)
let ajoute_cons (p1,n1,_ as c1) (p2,n2,_ as c2) syst =
  match (intersect p1 n2, intersect p2 n1) with
    | ([x],[]) -> ajoute (union (except x p1) p2, union n1 (except x n2), D(c1,c2)) syst
    | ([],[x]) -> ajoute (union p1 (except x p2), union (except x n1) n2, D(c1,c2)) syst
    | _        -> syst
;;

(* ajoute  l tous les consensus entre c et une clause du systme *)
let rec ajoute_consensus c syst l = match syst with
  | [] -> l
  | c1::suite -> ajoute_consensus c suite (ajoute_cons c c1 l)
;;

(* clt un systme par consensus *)
let rec cloture s =
  let rec consensus s1 s2 = match s1 with
    | [] -> s2
    | c::suite -> consensus suite (ajoute_consensus c s s2)
  in
  let s' = simplifie(consensus s s) in
  if (subtract s s' = []) && (subtract s' s = []) then s else cloture s'
;;

(* explique les tapes d'une dduction *)
let explique clause =
  let vus = ref [] in
  let rec expl (p,n,r as c) =
  if not(mem c !vus) then begin
    match r with
      | H -> print_string "J'ai l'hypothse : ";
	     print_string (string_of_clause c);
	     print_newline();
	     vus := c :: !vus
      | D((p1,n1,r1 as c1),(p2,n2,r2 as c2)) ->
	     if r1 = H then (expl c2; expl c1) else (expl c1; expl c2);
	     print_string "De ";  print_string(string_of_clause c1);
	     print_string " et "; print_string(string_of_clause c2);
	     print_string " je dduis : "; print_string(string_of_clause c);
	     print_newline();
	     vus := c :: !vus
  end
  in expl clause
;;
                          (* +----------+
                             |  essais  |
                             +----------+ *)


(* construction de clauses *)
let prefix =>   a b = ([b],  [a],   H)  (* a => b         *)   
and prefix =>~  a b = ([],   [a;b], H)  (* a => non b     *)
and prefix ~=>  a b = ([a;b],[],    H)  (* non a => b     *)
and prefix ~=>~ a b = ([a],  [b],   H)  (* non a => non b *)
;;

(* devoirs et mauvaises notes *)
let syst = cloture ["d"  =>  "c"; "c"  =>~ "m"; "d" ~=>~ "m"]
in explique (hd(tl syst));;

(* club cossais *)
let syst = cloture ["e" ~=>  "r";
		    "r"  =>  "k";
		    "m"  =>~ "d";
		    "d"  =>  "e"; "e" => "d";
		    "k"  =>  "e"; "k" => "m";
		    "e"  =>  "k"]
in explique (hd syst);;
